/*
 * KmemStat v1.0 (C) Innosecc.com 2009
 *
 * This tool is used to analyze the running kernel and check for insinuations such as the following...
 *
 * kernel rootkits that directly & indirectly modify the syscall table
 * internal kernel function hijacking
 * on-the-fly modification to kernel code through /dev/kmem and aspects of /dev/mem
 * 
 * Features:
 *
 * 1. KmemStat displays the modified code against its original code
 * 2. KmemStat can analyze the running kernel code by symbol, memory-address range, or the entire kernel
 * 3. KmemStat can dump raw instruction bytes from anywhere in the kernel image or kernel memory
 * 4. KmemStat runs not only on Linux but also FreeBSD.
 * 5. KmemStat is Runtime Memory barrier patching conscious, it looks at the .altinstructions and .alt_replace
 * sections of vmlinux and learns the instruction patching for your CPU, therefore it will not confuse
 * runtime CPU optimization patching, with a kernel rootkit =)
 * 6. KmemStat is portable between x86 and x86_64 
 *
 * KmemStat requires you have access to your uncompressed kernel image that corresponds to the kernel
 * running in memory.
 * 
 * In Linux this is usually /usr/src/linux/vmlinux, and in FreeBSD it might be /boot/kernel.i386/kernel
 * or /boot/kernel.amd64/kernel or /boot/kernel/kernel
 *
 * To compile:
 * Linux    gcc elf.c -o elfstat
 * FreeBSD  gcc elf.c -o elfstat -lkvm
 *
 * ElfStat (C) 2008 Ryan O'Neill <ryan@bitlackeys.com> 
 * 
 */


#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <sys/types.h>
#include <string.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <getopt.h>
#include <string.h>
#include <sys/utsname.h>
#include <elf.h>
#include <sys/mman.h>
#include <errno.h>

#ifdef __FreeBSD__
#include <kvm.h>
#include <nlist.h>
#endif

/* PAGE_SIZE may be 8k on IA64 */

#define PAGEUP(x) (x + PAGE_SIZE - (x & (PAGE_SIZE - 1)))

#define PAGEDOWN(x) (x & ~(PAGE_SIZE - 1))


#define PAGE_SIZE 4096
#define BUFSZ 200
#define SYMBOLS "/proc/kallsyms" 
#define DELIM ": "
 
#define ALTINSTR ".altinstructions"
#define REPLACE ".altinstr_replace"
#define EX_TABLE "__ex_table"
#define SMP_ALTINSTR ".smp_altinstructions"

#define KMEM "/dev/kmem"
#define KMT_MEM "/dev/kmt_mem" /* Our custom alternative */

#ifdef x86_64 
	typedef Elf64_Ehdr ElfX_Ehdr;
	typedef Elf64_Phdr ElfX_Phdr;
	typedef Elf64_Shdr ElfX_Shdr;
	typedef Elf64_Addr ElfX_Addr;
	typedef Elf64_Off  ElfX_Off;
	typedef Elf64_Word ElfX_Word;
#else
	typedef Elf32_Ehdr ElfX_Ehdr;
	typedef Elf32_Phdr ElfX_Phdr;
        typedef Elf32_Shdr ElfX_Shdr;
	typedef Elf32_Addr ElfX_Addr;
        typedef Elf32_Off  ElfX_Off;
	typedef Elf32_Word ElfX_Word;

#endif
#ifdef x86_64
typedef unsigned long kmem_off_t;
#else
typedef unsigned long long kmem_off_t;
#endif

unsigned int mismatched_bytes;
/* Linux runtime patching baby!
 * Alternative instructions
 */
struct alt_instr {
        unsigned char *instr;              /* original instruction */
        unsigned char *replacement;     /* replacement instruction */
        unsigned char cpuid;              /* cpuid bit set for replacement */
        unsigned char instrlen;           /* length of original instruction */
        unsigned char  replacementlen;     /* length of new instruction, <= instrlen */
        unsigned char  pad1;
};

struct opts {
       char verbose;
       char shdr;
       char phdr;
       char all;
       char idt;
       char mem;
       char sym;
       char hex; 
       unsigned long int bytes;
       char sys;
       char sysmap;
       char kern;
       char altinstr;
       char view_alts;
       char debug;
       char kmt_mem;
       unsigned long addr;

} opts;

struct text_segment {
	ElfX_Addr vaddr; 
	ElfX_Word memsz;
	ElfX_Word filesz;
	ElfX_Off ex_table_off;
	ElfX_Addr ex_table_vaddr;
	ElfX_Off rodata_off;
	ElfX_Addr rodata_vaddr;
	ElfX_Word rodata_size;
	ElfX_Off poff;

	ElfX_Addr text_section_vaddr;
	ElfX_Off text_section_off;

} ts; 

struct kernel_image
{
	unsigned int totlen;
	char *path;
} kernel_image;


/* SMP LOCK PREFIX crap */
unsigned char smp_orig = 0x90;
unsigned char smp_repl = 0xf0;

unsigned int gtb;
int parse_smp_alternatives = 0;

unsigned long next_symaddr;
int kfd, vfd;
struct stat gstk, gstv;
struct stat st;
char *altinstr;

struct data_seg {
	Elf32_Addr vaddr;
	Elf32_Word memsz;
	Elf32_Word filesz;
	Elf32_Addr vaddr_end;
	Elf32_Off poff;
} ds;

/* Information for our custom runtime patching analysis */
struct runtime_patch_info
{
	unsigned int resolved_patches;

	ElfX_Addr alt_vaddr;
	ElfX_Word alt_size;
	ElfX_Off alt_off;
	
	ElfX_Addr alt_repl_vaddr;
	ElfX_Word alt_repl_size;
	ElfX_Off alt_repl_off;

	ElfX_Addr instruction_vaddr;
	ElfX_Addr replacement_vaddr;
	ElfX_Off instruction_offset;
	ElfX_Off replacement_offset;
	
	uint8_t *replacement_pointer;
	uint8_t *instruction_pointer;
	char instruction[20];
	char replacement[20];
	unsigned int alt_count;
	
	/* SMP stuff */
	ElfX_Addr alt_smp_vaddr;
	ElfX_Addr alt_smp_repl_vaddr;
	ElfX_Word alt_smp_size;
	ElfX_Word alt_smp_repl_size;
} runtime_patch;
	
struct patched_area
{
	ElfX_Addr address;
	int bytes;
};


/* skip this instruction segment from address to address + bytes */
//struct rtmbp_instr *skip_i_seg;

/* Locate where the text segment starts from the program header (We use shdr in Fbsd though) */
void parse_elf_header(char *vmlinux)
{ 
	int fd, i = 0;
	char *mem, *StringTable;
	int found_alt = 0, found_alt_smp = 0;
       
	ElfX_Ehdr *ehdr;
	ElfX_Phdr *phdr;
	ElfX_Shdr *shdr;
	
	if (opts.verbose)
		printf("[+] Verbose ELF info:\n");

	if ((fd = open(vmlinux, O_RDONLY)) < 0)
	{
		perror("[TOOL_FAILURE] open");
		exit(-1);
	}
	if (fstat(fd, &st) < 0)
	{
		perror("[TOOL_FAILURE] fstat");
		exit(-1);
	}
	mem = mmap(NULL, st.st_size, PROT_READ, MAP_PRIVATE, fd, 0);
 	if (mem == MAP_FAILED)
	{
		perror("[TOOL_FAILURE] mmap");
		exit(-1);
	}
	close(fd);

	/* Build ELF header */
	ehdr = (Elf32_Ehdr *)mem;
 	
	/* Verify ELF Magic */ 
        if (ehdr->e_ident[0] != 0x7f && strcmp(&ehdr->e_ident[1], "ELF"))
	{
		printf("[TOOL_FAILURE]\n%s it not an elf file\n",vmlinux);
		exit(-1);
	} 

	
	/* Get Text and Data segment info */
	phdr = (ElfX_Phdr *)(mem + ehdr->e_phoff);
	for (i = 0; i < ehdr->e_phnum; i++)
		if (phdr[i].p_type == PT_LOAD && phdr[i].p_flags == (PF_R | PF_X))
		{
			/* TEXT */
		 	ts.memsz = phdr[i].p_memsz;
			ts.filesz = phdr[i].p_filesz;
			ts.poff = phdr[i].p_offset; 
			ts.vaddr = phdr[i].p_vaddr; 
			
			/* DATA */
			ds.memsz = phdr[i + 1].p_memsz;
			ds.filesz = phdr[i + 1].p_filesz;
			ds.poff = phdr[i + 1].p_offset;
			ds.vaddr = phdr[i + 1].p_vaddr;
			
			kernel_image.totlen = (ds.poff + ds.filesz) - ts.poff;
			break;
		} 

	if (ehdr->e_shstrndx == SHN_UNDEF)
        {
                printf("[TOOL_FAILURE]\nNo section name string header, cannot verify .__ex_table\n");
                exit(-1);
        }

        shdr = (Elf32_Shdr *) (mem + ehdr->e_shoff);
        StringTable = &mem[shdr[ehdr->e_shstrndx].sh_offset];
	for (i = 0; i < ehdr->e_shnum; i++)
	{
		if (strcmp(&StringTable[shdr[i].sh_name], ".rodata") == 0)
		{
			ts.rodata_off = shdr[i].sh_offset;
			ts.rodata_vaddr = shdr[i].sh_addr;
			ts.rodata_size = shdr[i].sh_size;
			if (opts.verbose)
				printf("[+] Found .rodata at vaddr: 0x%x offset 0x%x\n", ts.rodata_vaddr, ts.rodata_off);
			break;
		}
	}

/* In FreeBSD we use end of .text as the threshold instead of beginning of ex_table */
#ifdef __FreeBSD__
	for (i = 0; i < ehdr->e_shnum; i++)
        {
                if (strcmp(&StringTable[shdr[i].sh_name], ".text") == 0)
                {
                        ts.text_section_off = shdr[i].sh_offset;
			ts.text_section_vaddr = shdr[i].sh_addr;
                        if (opts.verbose)
                                printf("[+] Found .text at vaddr: 0x%x offset 0x%x\n", ts.text_section_vaddr, ts.text_section_off);
                        break;
                }
        }
#endif

#ifdef __linux__
 
	for (i = 0; i < ehdr->e_shnum; i++)
        {
		if (strcmp(&StringTable[shdr[i].sh_name], EX_TABLE) == 0)
		{
			ts.ex_table_off = shdr[i].sh_offset;
			ts.ex_table_vaddr = shdr[i].sh_addr;
			if (opts.verbose)
				printf("[+] Found __ex_table at vaddr: 0x%x offset 0x%x\n", ts.ex_table_vaddr, ts.ex_table_off);
			break;
		}

	}
	 
		  
	  /* On some kernels .smp_lock section (And probably others) */
	  /* appear to be .altinstructions based on attributes, so I'm */
	  /* using the string table to confirm the name of the section */
	  
       	  shdr = (ElfX_Shdr *)(mem + ehdr->e_shoff);
	  StringTable = &mem[shdr[ehdr->e_shstrndx].sh_offset];
    
	  for (i = 0; i < ehdr->e_shnum; i++)
	  {	
		    	if (strcmp(&StringTable[shdr[i].sh_name], ALTINSTR) == 0)
		    	{
			    	runtime_patch.alt_vaddr = shdr[i].sh_addr;
				runtime_patch.alt_size = shdr[i].sh_size;
				found_alt++;
		    	} 
			else
			if (strncmp(&StringTable[shdr[i].sh_name], ".altinstr_repl", 14) == 0)
			{
				 runtime_patch.alt_repl_vaddr = shdr[i].sh_addr;
                                 runtime_patch.alt_repl_size = shdr[i].sh_size;
				 found_alt++;
			}
	   }	
	   if (opts.verbose && found_alt == 2)
	   {
	   	printf("[+] Found .altinstructions\t0x%x\t%d bytes\n", runtime_patch.alt_vaddr, runtime_patch.alt_size);
		printf("[+] Found .altinstr_replace\t0x%x\t%d bytes\n", runtime_patch.alt_repl_vaddr, runtime_patch.alt_repl_size);
	   }

	if (!found_alt)
		printf("[!] Unable to locate alternative instructions, kernel malware detection may be unreliable\n");
	 
	for (i = 0; i < ehdr->e_shnum; i++)
	{
		if (strcmp(&StringTable[shdr[i].sh_name], SMP_ALTINSTR) == 0)
		{
			runtime_patch.alt_smp_vaddr = shdr[i].sh_addr;
			runtime_patch.alt_smp_size = shdr[i].sh_size;
			found_alt_smp++;
		}
		else
		if (strncmp(&StringTable[shdr[i].sh_name], ".smp_altinstr_repl", 18) == 0)
		{
			runtime_patch.alt_smp_repl_vaddr = shdr[i].sh_addr;
			runtime_patch.alt_smp_repl_size = shdr[i].sh_addr;
			parse_smp_alternatives = 1;
			found_alt_smp++;
			break;
		}
	}
	if (opts.verbose && found_alt_smp == 2)
	{
		printf("[+] Found .smp_altinstructions\t0x%x\t%d bytes\n", runtime_patch.alt_smp_vaddr, runtime_patch.alt_smp_size);
		printf("[+] Found .smp_altinstr_replace\t0x%x\t%d bytes\n", runtime_patch.alt_smp_repl_vaddr, runtime_patch.alt_smp_repl_size);
	}
#endif 

	/* No real point in returning it since its global */		
	munmap(mem, st.st_size);
}
 
/* dump instruction bytes from both vmlinux and memory */
void hexdump(void *mem, int len, char type)
{	
	int i;
	unsigned char *p = mem;
	
	if (type == 1)
		printf("printing first %d bytes of .text segment from /dev/kmem\n",len);
	else
		printf("printing first %d bytes of .text segment from kernel image\n",len);

	for (i = 0; i < len; i++)
		printf("\\x%.2x",*p++);
	puts("\n");
}

void hexdump_alt_instr(void *mem, unsigned int off, unsigned int len, char type)
{
	int i;
	unsigned char *p = mem + off;
	
	if (type == 1)
		printf("Printing %d bytes of .altinstructions section\n",len); 
	else
		printf("Printing %d bytes of .altinstr_replace section\n",len); 
	for (i = 0; i < len; i++)
		printf("\\x%.2x", *p++);
	puts("\n");
}

int check_rodata(uint8_t *mem1, uint8_t *mem2, ElfX_Addr start_vaddr, ElfX_Addr end_vaddr, struct patched_area *patch)
{
	uint8_t *p, *q;
	ElfX_Addr vaddr = start_vaddr;
	unsigned int alt_resolved = 0, mismatch = 0, check = 0, c = 0, j;

	for (p = mem1, q = mem2; vaddr < end_vaddr; vaddr++, c++)
	{
		if (opts.debug)
			printf("Comparing %.2x and %.2x\n", *p, *q);
		
		if (*p != *q)
		{
			#ifdef __linux__
                        for (j = 0; j < runtime_patch.alt_count; j++)
                        {
                                if (vaddr == patch[j].address)
                                {
                                        p += patch[j].bytes - 1;
                                        q += patch[j].bytes - 1;
                                        vaddr += patch[j].bytes - 1;
                                        alt_resolved++;
                                        check++;
                                        break;
                                }
                        }
			#endif
			if (!check)
				mismatch++;
			if (opts.verbose && !check)
                                printf("[0x%x] Bytes do not match: [Original] x%.2x [Modified] x%.2x\n",vaddr, *q, *p);
		}
		end:
		check = 0;
		p++;
		q++;
	}

	/* cmp_text_seg() will be called before check_rodata() so mismatched_bytes may already be non zero */
	mismatched_bytes += mismatch;
	gtb += c;
}
unsigned long int cmp_text_seg(uint8_t *mem1, uint8_t *mem2, unsigned long int size, unsigned long vaddr, struct patched_area *patch)
{
	unsigned int i, c = 0, t = 0, j;
	unsigned char *p, *q;
	unsigned int debug = 0;
	unsigned int alt_resolved = 0, check = 0, mismatch = 0;
	ElfX_Addr vaddr_end;

	p = (uint8_t *)mem1;
	q = (uint8_t *)mem2;
	
	if (opts.verbose)
		printf("[+] Start vaddr: 0x%x End Vaddr: 0x%x size: 0x%x\n", vaddr, vaddr+size, size);
	
	/* Set last vaddr value */
	vaddr_end = vaddr + size;
	
	/* Loop through and compare the kernel image to kernel memory */
	/* and perform runtime patch analysis as needed */
	
#ifdef __linux__
	if (opts.mem || opts.sym && ts.vaddr >= ts.ex_table_vaddr)
		printf("[+] Vaddr 0x%x Exceeding .text - Results may be unreliable in Linux\n", ts.vaddr);
#endif
	for (; vaddr < vaddr_end; vaddr++, c++)
	{
		/* Lets not exceed beyond .text */ 	
#ifdef __linux__
		if (opts.all)
			if (vaddr >= ts.ex_table_vaddr)
				break;
#else
		if (vaddr >= ts.vaddr + ts.text_section_off)
			break;
#endif	
		if (opts.debug)
			printf("Comparing %.2x and %.2x\n", *p, *q);
		
		if (*p != *q)
		{
#ifdef __linux
			if (*p == smp_orig && *q == smp_repl)
				goto end;
#endif
			#ifdef __linux__
			for (j = 0; j < runtime_patch.alt_count; j++)
			{
				if (vaddr == patch[j].address)
				{
					p += patch[j].bytes - 1;
					q += patch[j].bytes - 1;
					vaddr += patch[j].bytes - 1; 
					alt_resolved++;
					check++;
					break;
				}
			}
			#endif
			if (!check)
				mismatch++;
	
			if (opts.verbose && !check) 
				printf("[0x%x] Bytes do not match: [Original] x%.2x [Modified] x%.2x\n",vaddr, *q, *p);		
		}	 
		end:
		check = 0;
		p++;
		q++; 
		
	}
	
	runtime_patch.resolved_patches = alt_resolved;

	if (opts.debug)
		printf("In comparing text segment, kmemstat skipped over %d instruction bytes\n",debug);
	
	mismatched_bytes = mismatch;
	gtb = c;
	return mismatch;
}

#ifdef __FreeBSD__
unsigned long get_symbol_fbsd(char *name)
{ 
	char err[255];
	kvm_t *kd;
	struct nlist nl[] = { { NULL }, { NULL }, };

	nl[0].n_name = name;

	kd = kvm_openfiles(NULL, NULL, NULL, O_RDONLY, err);
	if (!kd)
	{
		printf("get_symbol_fbsd: %s\n",err);
		return -1;
	}

	if (kvm_nlist(kd, nl) < 0) 
	{
		printf("get_symbol_fbsd: %s\n",kvm_geterr(kd));
		return -1;
	}
            
	if(nl[0].n_value) 
		return nl[0].n_value;
	else
	return -1;

}
#endif

unsigned long get_symbol(char *name, char *syms)
{
	FILE *fd;
	char buf[BUFSZ], symbol_s[BUFSZ], type;
	int i;
	unsigned long address;
	char tmp[BUFSIZ], tmp_type;
	
	if (!syms)
	{
		if ((fd = fopen(SYMBOLS, "r")) == NULL)
		{ 	
			perror("kallsyms fopen");
			return -1;
		}
	}
	else
		if ((fd = fopen(syms, "r")) == NULL)
		{ 
			perror("system.map fopen");
			return -1;
		}
	
	for (; !feof(fd); )
	{ 
		bzero(symbol_s, BUFSZ);  
		/* get symbol address for symbol we want */
		fscanf(fd, "%x %c %s", &address, &type, &symbol_s);
		if (strcmp(symbol_s, name) == 0)
		{
			/* get the next symbol address as well so we have a range of addresses */
			/* this is not really the best method of figuring out when the instructions */
			/* for one symbol end at all, but will use it for now */
			fscanf(fd, "%x %c %s", &next_symaddr, &tmp_type, &tmp); 
			fclose(fd);
			return address;
		} 
		/* kallsyms isn't a regular file, meaning there is no EOF */
		if (!strcmp(symbol_s, ""))
			break;
	} 
	return 0;
} 

unsigned int round_up(unsigned int x, unsigned int y) 
{
	return (x + y - 1) / y * y; 
}

int main(int argc, char **argv)
{ 

	opts.phdr = 0;
	opts.shdr = 0;
 
 	opts.all = 0;
	opts.idt = 0;
	opts.mem = 0;
	opts.sym = 0;
	opts.hex = 0; 
	opts.sys = 0;
	opts.sysmap = 0;
	opts.kern = 0;
	opts.view_alts = 0;
	opts.debug = 0;
	opts.kmt_mem = 0;
        opts.altinstr = 1;
	
 	unsigned long int count, cmp, hexbytes;
 	unsigned char  *pg2, *pg3;
	unsigned long int offset, addr;
	unsigned long symaddr = 0;
        char fbsd = 0;   	
	extern char *optarg;
	char targetOS[512];
	struct utsname uts;

	char *p, path[255], *op, *q; 
	char *sysmap, *symbol;
 	int i = 0, j = 0;
	int opt_index = 0, c;
	int total = 0;
	unsigned int sr;
	
	struct rtmbp_instr *skip_i_seg;
	struct smp_instr *skip_smp_seg;
	struct patched_area *patch;
			
	ElfX_Addr vaddr_end;
	char kmem_path[255];
	
	if (argc < 4)
	{ 
		usage:
		printf("Kmemstat 1.0 - (C) Innosecc.com 2009 <ryan@innosecc.com>\n");
		printf("Usage: %s -k <kernel image> <options>\n", argv[0]); 
		printf("[-k] To specify kernel image, i.e /usr/src/linux/vmlinux or in FreeBSD /boot/kernel/kernel\n");
		printf("[-a] To analyze all kernel code within .text section (covers most things)\n");
		printf("[-m] <address>:<bytes> To check a specific memory range from address to address+bytes\n");
		printf("[-s] symbol[:<system.map>] To analyze the code for a specific kernel symbol, uses /proc/kallsyms by default\n");
		printf("[-x] <bytes> To dump the first X amount of bytes of both vmlinux and kernel memory .text segment to stdout\n");
		printf("[-v] Verbose is recommended, as it displays the actual code modifications\n");
		printf("[-S] Analyze the code for every system call, report any modifications (Unfinished)\n");
		printf("[-o] Show CPU optimization runtime patching information from .altinstructions/.altinstr_replace\n");
		printf("[-e] Show ELF segment/section virtual addresses and offsets, useful before using '-m'\n");
		printf("[-M] Use custom /dev/kmt_mem (instead of /dev/kmem)\n");
		printf("[-d] Debug - Fun for the family & friends\n");
		printf("[-h] Help (This menu)\n");
		printf("Note: For option '-a', elfstat scans memory through the entire executable portion of .text, but ot the entire segment;\n");
		printf("To reach certain symbols that are closer to the data segment i.e sys_call_table, use the '-s' option\n");
		printf("Note: For option '-s', a path to System.map is optional, /proc/kallsyms is used by default in linux,\n"); 
		printf("FreeBSD (conveniently) doesn't require any files to retrieve symbol information\n");
		printf("Example 1: ./elfstat -k /usr/src/linux/vmlinux -m 0xc071d2d0:100 -v\n");
		printf("Example 2: ./elfstat -k /usr/src/linux/vmlinux -s sys_getdents64\n"); 
		printf("Example 3: ./elfstat -k /boot/boot/kernel -m 0x80232773:75\n");
		printf("Example 4: ./elfstat -k /usr/src/linux/vmlinux -e\n");
		exit(0);
	} 
        
	
	if (getuid() != 0)
	{
		printf("%s must be run as root!\n", argv[0]);
		exit(0);
	}
 
		
	if (uname(&uts) < 0)
	{	
		printf("[TOOL_FAILURE] uname() %s\n", strerror(errno));
		exit(-1);
	}


	snprintf(targetOS, sizeof(targetOS), "%s %s %s", uts.sysname, uts.release, uts.version);

	while ((c = getopt(argc, argv, "hvaim:s:x:Sk:odeM")) != -1) 
	{
		switch(c) 
		{
			case 'a':
				opts.all++;
				total++;
				break;
			case 'i':
				opts.idt++; 
				total++;
				break;
			case 'm':
				opts.mem++; 
                                if (!optarg)
					goto usage; 
				q = strchr(optarg, 0x3A);
				if (!q)
				{
					printf("\n-m requires two arguments delimited by ':'\n\n");
					goto usage;
				}
				op = strtok(optarg, DELIM);
				opts.addr = strtoul(op, NULL, 16); 

				op = strtok(NULL, DELIM);
				if (!op)
					goto usage;
				opts.bytes = (unsigned long int)strtoul(op, NULL, 10);

				total++;
				break;
			case 's':
				opts.sym++; 
 				if (!optarg)
					goto usage;
				q = strchr(optarg, 0x3A);
				if (!q)
				       symbol = strdup(optarg);
				else
				{
					op = strtok(optarg, DELIM);
					symbol = strdup(op);

					op = strtok(NULL, DELIM);
					if (!op)
						goto usage; 
					sysmap = strdup(op);
					opts.sysmap = 1;

				}
				total++;
				break;
			case 'v':
				opts.verbose++; 
				break;
			case 'x': 
				if (!optarg)
					goto usage;
				opts.hex++;
				hexbytes = (unsigned long int)strtoul(optarg, NULL, 10);
				break;
			case 'S':
				opts.sys++;
				break;
			case 'k': 
				opts.kern++; 
				if (!optarg)
					goto usage;
				memcpy((char *)path, (char *)optarg, sizeof(path) - 1);
				break;
			case 'o':
				#ifdef __FreeBSD__
				printf("This option does not apply since FreeBSD (Apparently) doesn't do runtime patching\n");
				exit(-1);
				#endif
				opts.view_alts++;
				break;
			case 'd':
				opts.debug++;
				break;
			case 'e':
				/* Really just setting verbose here */
				opts.verbose++;
				break;
			case 'M':
				opts.kmt_mem++;
				break;
	 		default:
				puts("[!] Unknown option used\n");
				goto usage;
		}
	}
 	
	if (opts.mem && opts.sym)
	{
		printf("Options -m and -s cannot be used together.\n");
		exit(0);
	}

	if (!opts.kern)
	{	
		printf("Must specify a kernel image, i.e /usr/src/linux/vmlinux\n");
		goto usage;
	}
	
	
	#ifdef __FreeBSD__
		opts.altinstr = 0;
	#endif
	opts.phdr = 1;
	parse_elf_header(path);
        
	if (opts.mem)
	{
		unsigned long int toff;
		toff = ts.vaddr + ts.memsz;

		if ((opts.bytes + ts.vaddr) > toff)
		{
			printf("%d bytes exceeds the .text segment\n", opts.bytes);
			exit(-1);
		}
	} 
 
	if (opts.hex)
	{
		unsigned long int toff;
		toff = ts.vaddr + ts.memsz;

		if ((hexbytes + ts.vaddr) > toff)
		{
			printf("%d bytes exceeds the .text segment\n", hexbytes);
			exit(-1);
		}
	}


	if(opts.verbose)
		printf("[+] %s's .text segment starts at 0x%x, and its %d bytes\n", path, ts.vaddr, ts.filesz);
	
	
	/* Open /dev/kmem */
	if (opts.kmt_mem)
		strcpy(kmem_path, KMT_MEM);
	else
		strcpy(kmem_path, KMEM);

	if((kfd = open(kmem_path, O_RDONLY)) == -1)
	{
		printf("[TOOL_FAILURE] open %s %s\n",kmem_path, strerror(errno));
		exit(-1);
	}
	
	/* Open /usr/src/linux/vmlinux or the likes */
	if((vfd = open(path, O_RDONLY)) == -1)
	{
		perror("[TOOL_FAILURE] open kernel image");
		exit(-1);
	}
	
	
	unsigned long kb; // kmem bytes

	/* Lets do stack allocation instead of heap, as it may improve performance */
	/* slightly. */
	unsigned char *kmem = alloca(ts.memsz);
	unsigned char *kimage;

	/* In some Linux distros its necessary to lseek64/read /dev/kmem, and mmap() does not work */
	/* As a result we will use lseek64/read in FreeBSD as well, rather than using mmap() */

	/* Find text segment */
#ifdef __linux__
	if (lseek64(kfd, (kmem_off_t)ts.vaddr, SEEK_SET) == -1) 
	{
		perror("lseek kmem");
		exit(-1);
	}
#else
	if (lseek(kfd, (kmem_off_t)ts.vaddr, SEEK_SET) == -1)
        {
                perror("lseek kmem");
                exit(-1);
        }
#endif
	/* Read entire text segment onto our stack pointed to by kmem */
	if((kb = read(kfd, kmem, ts.memsz)) < 0)
	{
		perror("read kmem");
		exit(-1);
	} 
	
	/* Map uncompressed kernel image into memory */
  	kimage = mmap(NULL, PAGEUP(kernel_image.totlen), PROT_READ, MAP_PRIVATE, vfd, PAGEDOWN(ts.poff));
 	if (kimage == MAP_FAILED)
	{
		perror("kernel image mmap");
		exit(-1);
	}

	
	/*
	  struct alt_instr { 
	  	u8 *instr; 	        original instruction 
		u8 *replacement;	new instruction
 	 	u8  cpuid;	 	cpuid bit set for replacement 
	 	u8  instrlen;		length of original instruction 
	 	u8  replacementlen; 	length of new instruction, <= instrlen 
	}; 
	elfstat has to parse runtime memory barrier patching structs in vmlinux
	/usr/src/linux/arch/x86/kernel/alternative.c 

	*/
	

	#ifdef __linux__
	
	/* for altinstructions offsets */
	runtime_patch.alt_off = runtime_patch.alt_vaddr - ts.vaddr;
	
	/* for altinstr_replacement offsets */
	runtime_patch.alt_repl_off = runtime_patch.alt_repl_vaddr - ts.vaddr;

	/* Determine the number of alt_instr structs there are in the .altinstructions section */
	runtime_patch.alt_count = round_up(runtime_patch.alt_size, sizeof(struct alt_instr)) / sizeof(struct alt_instr);

	/* Alternative instruction pointer, is now pointing to .altinstructions in vmlinux */
	struct alt_instr *ap = (struct alt_instr *)&kimage[runtime_patch.alt_off];
	
	if((patch = malloc(runtime_patch.alt_size * sizeof(struct patched_area))) == NULL)
	{
		perror("[TOOL_FAILURE] malloc");
		exit(-1);
	}
		
	for (i = 0; i < runtime_patch.alt_count; i++, ap++) 
	{	
		/* We're only interested in the .text segment */
		if ((unsigned long)ap->instr > ts.vaddr + ts.memsz)
			continue;

	 	if (opts.view_alts)
		{
			printf("[+] instr (located at) %x replacement (located at) %x cpuid = %x instrlen = %d replacementlen = %d pad1 = %d\n", 
			ap->instr, ap->replacement, ap->cpuid, ap->instrlen, ap->replacementlen, ap->pad1);
		} 
	
		/* Patch location and size */
        	patch[i].address = (unsigned long)ap->instr;
		patch[i].bytes = (unsigned int)ap->instrlen; 
		
		
		/* Get old instruction info */
		runtime_patch.instruction_vaddr = (unsigned long)ap->instr;
		runtime_patch.instruction_offset = runtime_patch.instruction_vaddr - ts.vaddr;
		runtime_patch.instruction_pointer = (unsigned char *)&kimage[runtime_patch.instruction_offset];

		/* Get new instruction info */
		runtime_patch.replacement_vaddr = (unsigned long)ap->replacement;
		runtime_patch.replacement_offset = runtime_patch.replacement_vaddr - ts.vaddr;
		runtime_patch.replacement_pointer = (unsigned char *)&kimage[runtime_patch.replacement_offset];
			

		/* Display old and new instruction for '-o' option */
		if (opts.view_alts)
		{
			printf("[+] Runtime Patch-> Original instructions:\n");
			for (j = 0; j < ap->instrlen; j++)
				printf("\\x%.2x", runtime_patch.instruction_pointer[j]);
			puts("\n");

			printf("[+] Runtime Patch-> Replacement instructions:\n");
			for (j = 0; j < ap->replacementlen; j++)
				printf("\\x%.2x", runtime_patch.replacement_pointer[j]);

			puts("\n"); 
		}

	}
	
#endif


	if (opts.hex) 
	{       
		if (total)
		{
			printf("--hex-dump must be run without any other options\n");
			goto usage;
		}
		hexdump(kmem, hexbytes, 1);
		hexdump(kimage, hexbytes, 0); 
		goto done;
	}
	
	if (opts.all)
	{
		cmp = 0; 
		#ifdef __linux__
			cmp = cmp_text_seg(kmem, kimage, ts.ex_table_off, ts.vaddr, patch);
			
			/* Must leave this out for now, turns out .rodata is NOT rodata. Seems to be patched in certain locations */
			//check_rodata(kmem, kimage, (ts.vaddr + ts.rodata_off), (ts.vaddr + ts.rodata_off + ts.rodata_size), patch);
		#else
			cmp = cmp_text_seg(kmem, kimage, ts.text_section_off, ts.vaddr, NULL);
			//check_rodata(kmem, kimage, (ts.vaddr + ts.rodata_off), (ts.vaddr + ts.rodata_off + ts.rodata_size), NULL);
		#endif 
		
		unsigned long txend;
		
#ifdef __FreeBSD__
		txend = ts.vaddr + ts.filesz;
#else
		txend = ts.ex_table_vaddr;
#endif
		if (cmp)
		{ 
			printf("[ALERT_TO_DAEMON]\n- KmemStat Results [%s] -\n\n"
			       "- Total number of instruction bytes analyzed: %d -\n"
			       "- Total number of resolved runtime patches:   %d -\n"
			       "- Total number of mismatched instructions:    %d -\n"
			       "- Kernel memory status: Malicious code modifications detected between 0x%x and 0x%x\n"
			       "- Please use '-v' option to see the actual code changes\n\n",targetOS, gtb, runtime_patch.resolved_patches, mismatched_bytes, ts.vaddr, txend);
		}
		else

			printf("\n- KmemStat Results [%s] -\n\n"
			       "- Total number of instruction bytes analyzed: %d -\n"
			       "- Total number of resolved runtime patches:   %d -\n"
                               "- Total number of mismatched instructions:    %d -\n"
			       "- Kernel memory status: No malicious code modifications detected between 0x%x and 0x%x\n\n",targetOS, gtb, runtime_patch.resolved_patches, mismatched_bytes, ts.vaddr, txend);
		goto done;
	}
        
	if (opts.mem)
	{
		vaddr_end = ts.vaddr;
		vaddr_end += ts.filesz; 
 
		if (opts.addr > vaddr_end) 
		{
			printf("[+] %x is an address outside of the .text segment, exiting.\n", opts.addr);
		 	goto done;
		} 

		offset = (unsigned int)(unsigned long)opts.addr - (unsigned long)ts.vaddr;

		if (opts.verbose)
			printf("%x is %d many bytes into the text segment\n",opts.addr, offset);
		cmp = 0;
		#ifdef __linux__
		printf("Calling cmp_text_seg\n");
		cmp = cmp_text_seg(&kmem[offset], &kimage[offset], opts.bytes, opts.addr, patch);
		#else
		cmp = cmp_text_seg(&kmem[offset], &kimage[offset], opts.bytes, opts.addr, NULL);
		#endif
		if (cmp)
		{
			printf("[+] KmemStat found modifications to the kernel somewhere between 0x%x and 0x%x, use -v to see code modifications\n",opts.addr, opts.addr + opts.bytes);
			goto done;
		}
		else
			printf("[+] KmemStat did not find any modifications to kernel memory between 0x%x and 0x%x\n",opts.addr, opts.addr + opts.bytes);
		goto done;
	}
 	
	if (opts.sym)
	{
		vaddr_end = ts.vaddr;
		vaddr_end += ts.filesz;
                 
		#ifdef __FreeBSD__
			symaddr = get_symbol_fbsd(symbol);
		#else
			if (opts.sysmap)
				symaddr = get_symbol(symbol, sysmap);
			else
				symaddr = get_symbol(symbol, NULL);  
	        #endif
		
		if (symaddr <= 0)
		{
			printf("[!] Unable to find symbol: %s\n", symbol);
			goto done;
		}
		if (opts.verbose)
			printf("[+] Symbol [%s]\t0x%x\n", symbol, symaddr);
 		
		if (symaddr > vaddr_end)
		{
			printf("[!] [%s <0x%x>] is outside of the .text segment, exiting\n", symbol, symaddr);
			goto done;
		} 
		offset = (unsigned int)(unsigned long)symaddr - (unsigned long)ts.vaddr; 

		next_symaddr -= 1;

		if (opts.verbose)
			#ifdef __FreeBSD__
			printf("\n[%s <0x%x>] is %u bytes into the text segment\n", symbol, symaddr, offset);
			#else
			printf("\n[%s <0x%x>] is %u bytes into the text segment, and ends at 0x%x\n",symbol, symaddr, offset, next_symaddr);
	 		#endif
	 	/* sr (symbol range) number of bytes between specified symbol and next symbol */
                
#ifdef __FreeBSD__
			/* Until I figure out how to use kvm lib all the way */
			/* I won't know how to determine where the next symbol starts */
			/* so right now lets just scan 256 bytes past each symbol */ 
			sr = 0x100;
			fbsd++; 
#else
 	              	sr = (unsigned int)(unsigned long)next_symaddr - symaddr;
#endif 
		
		cmp = 0;
#ifdef __linux__
		cmp = cmp_text_seg(&kmem[offset], &kimage[offset], sr, symaddr, patch);
#else
		cmp = cmp_text_seg(&kmem[offset], &kimage[offset], sr, symaddr, NULL);
#endif

		if (cmp)
		{
			printf("[+] [ALERT_TO_DAEMON] KmemStat found modifications to %s code somewhere between 0x%x and 0x%x, use -v to see code modifications\n", symbol, symaddr, next_symaddr);
			goto done;
		}
		else
			printf("[+] KmemStat did not find any modifications to [%s <0x%x>]\n", symbol, symaddr);
			free(symbol);
			if (opts.sysmap)
				free(sysmap);
		goto done;
	} 

	done:
	munmap(kimage, PAGEUP(ts.filesz));
	close(kfd);
	close(vfd); 
	exit(0);
}
