/*
 * snapmem (C) innosecc.com 2009
 * Snapmem takes a snapshot of the kernel code in memory and copies it
 * to disk as a raw binary, then creates a custom ELF header with a custom
 * section called ".text_snapshot" which other tools can read from for later
 * forensic analysis 
 * Author: Ryan O'Neill
 */

#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>
#include <stdlib.h>
#include "../common/common.h"

#include <errno.h>
#include <fcntl.h>
#include <string.h>
#include <stdarg.h>
#include <sys/mman.h>
#include <sys/stat.h>

void
exec_string (char *str, ...)
{
        char string[255];
        va_list va;

        va_start (va, str);
        vsnprintf (string, 255, str, va);
        va_end (va);

	printf("Adding ELF header:\n%s\n", string);
	
        system (string);
}

void read_kcore_elf(uint8_t *snapshot, unsigned int len)
{
	int fd;
	
	if ((fd = open("/proc/kcore", O_RDONLY)) == NULL)
	{
		printf("Failed to open /proc/kcore\n");
		exit(-1);
	}
	
	Elf32_Ehdr *ehdr;
	struct stat st;
	fstat(fd, &st); 
	read(fd, snapshot, len);
	
	ehdr = (Elf32_Ehdr *)snapshot;
	if (ehdr->e_type != ET_CORE)
	{
		printf("/proc/kcore does not appear to be of type ET_CORE\n");
		exit(0);
	}
	
	close(fd);
}
	
int main(int argc, char **argv)
{
	
	int use_kmt_mem = 0;
	int use_kcore = 0;
	char *outfile, kmem_path[16];
	unsigned char *snapshot;
	unsigned char *chunk;
	int fd, i;
	
	ElfX_Ehdr *ehdr;
	ElfX_Phdr *phdr;
	ElfX_Word textlen;
	ElfX_Addr text_start, text_end;

	if (argc < 2)
	{
		printf("Usage: %s <image_output> [--kmt_mem] [--kcore]\n", argv[0]);
		exit(-1);	
	}

	if ((outfile = strdup(argv[1])) == NULL)
	{
		printf("[TOOL_FAILURE] strdup() %s\n", strerror(errno));
		exit(-1);
	}

	
	if (argv[2])
	{
		if (strcmp(argv[2], "--kmt_mem") == 0)
			use_kmt_mem++;
		else
		if (strcmp(argv[2], "--kcore") == 0)
			use_kcore++;
	}

	if (use_kmt_mem)  
	{
                strcpy(kmem_path, "/dev/kmt_mem"); /* Custom /dev/kmem like device */
	}
	else 
	if (use_kcore)
	{
		strcpy(kmem_path, "/proc/kcore"); /* This is nice because it doesn't require vmlinux to get a live snapshot */
        }
	else
	{
                strcpy(kmem_path, "/dev/kmem"); /* This is default */
	}

        if ((kfd = open (kmem_path, O_RDONLY)) == -1)
        {
                printf("[TOOL_FAILURE] %s open %s\n", kmem_path, strerror(errno));
                exit(-1);
        }
	
	
	text_start = get_symbol("_text", "/proc/kallsyms");
	
	printf("Start of text: 0x%x\n", text_start);

	if (text_start == 0)
	{
		printf("Failure: Unable to lookup symbol: _text\n");
		exit(-1);
	}

	text_end = get_symbol("__end_rodata", "/proc/kallsyms");
	if (text_end == 0)
	{
		printf("Failure: Unable to lookup symbol: __end_rodata\n");
		exit(-1);
	}

	printf("End of text: 0x%x\n", text_end);
	
	/* Increase by another MB */
	text_end += 1000000;
	printf("+ 1MB = 0x%x (Some of our tools require a bit beyond)\n", text_end);

	textlen = text_end - text_start;

	if ((snapshot = (unsigned char *)malloc(textlen)) == NULL)
	{
		printf("[TOOL_FAILURE] malloc() %s\n", strerror(errno));
		exit(-1);
	}
	
	if (use_kcore)
		read_kcore_elf(snapshot, textlen);
	else
	if (readkmem(snapshot, text_start, textlen) == -1)
        {
                printf("[TOOL_FAILURE] readkmem failure\n");
                exit(-1);
       
	}

	if ((fd = open(outfile, O_CREAT | O_TRUNC | O_WRONLY, 0777)) == -1)
        {
	 	printf("[TOOL_FAILURE] open() %s\n", strerror(errno));
		exit(-1);
	 
	}
		
	printf("Writing file: %s\n", outfile);

	if (write(fd, snapshot, textlen) < 0)
	{
		printf("[TOOL_FAILURE] write() %s\n", strerror(errno));
		exit(-1);
	}

	exec_string("objcopy -I binary -O elf32-i386 -B i386 --rename-section .data=.text_snapshot %s %s", outfile, outfile);  
	
	exit(0);
}

